import glob
import json
import sys
import time, datetime
import pandas as pd
import numpy as np
import asyncio
import aiofiles
import aioitertools
from tqdm.auto import tqdm
import zipfile

from deep_orderbook.shapper import BookShapper

MARKETS = ["ETHBTC", "BTCUSDT", "ETHUSDT", "BNBBTC", "BNBETH", "BNBUSDT"]


class Replayer:
    def __init__(self, data_folder, date_regexp=''):
        self.data_folder = data_folder
        self.date_regexp = date_regexp

    def zipped(self):
        zs = sorted(glob.glob(f'{self.data_folder}/{self.date_regexp}*.zip'))
        for z in zs:
            fns = zipfile.ZipFile(z).namelist()
            print(fns)
            with zipfile.ZipFile(z) as myzip:
                with myzip.open(fns[0]) as myfile:
                    pass#print(myfile.read())

    def snapshots(self, pair):
        return sorted(glob.glob(f'{self.data_folder}/{pair}/{self.date_regexp}*snapshot.json'))

    def updates_files(self, pair):
        Bs = sorted(glob.glob(f'{self.data_folder}/{pair}/{self.date_regexp}*update.json'))
        return Bs

    def trades_file(self, pair):
        Ts = sorted(glob.glob(f'{self.data_folder}/{pair}/{self.date_regexp}*trades.json'))
        return Ts

    def book_updates_and_trades(self, pair):
        Bs = self.updates_files(pair)
        Ts = self.trades_file(pair) 
        return [(b, b.replace('update', 'trades')) for b in Bs if b.replace('update', 'trades') in Ts]

    def training_file(self, pair):
        BTs = sorted(glob.glob(f'{self.data_folder}/{self.date_regexp}*{pair}*ps.npy'))
        for fn_ps in BTs:
            fn_bs = fn_ps.replace('ps.npy', 'bs.npy')
            fn_ts = fn_ps.replace('ps.npy', 'time2level.npy')
            yield (fn_bs, fn_ps, fn_ts)

    def training_samples(self, pair):
        for (fn_bs, fn_ps, fn_ts) in self.training_file(pair):
            arr_books = np.load(fn_bs)
            arr_prices = np.load(fn_ps)
            arr_time2level = np.load(fn_ts)
            yield arr_books, arr_prices, arr_time2level

    @staticmethod
    def tradesframe(file):
        ts = pd.DataFrame(json.load(open(file)))
        if ts.empty:
            return ts
        ts = ts.drop(['M', 's', 'e', 'a'], axis=1).astype(np.float64)
        ts['t'] = 1 + ts['E'] // 1000
        ts['delay'] = ts['E'] - ts['T']
        ts['num'] = ts['l'] - ts['f'] + 1
        ts['up'] = 1 - 2*ts['m']
        ts.drop(['E', 'T', 'f', 'l', 'm'], axis=1, inplace=True)
        ts.set_index(['t'], inplace=True)
        return ts

    @staticmethod
    def sample(of_file):
        return json.load(open(of_file))[0]

    async def replayL2_async(self, pair, shapper):
        yield pair
        snapshotupdates = {}
        files = tqdm(self.snapshots(pair))
        for snap_file in files:
            try:
                snap = json.load(open(snap_file))
            except:
                continue
            snapshotupdates[snap['lastUpdateId']] = snap_file
        snapshotupdates = iter(sorted(snapshotupdates.items()))
        next_snap,snapshot_file = next(snapshotupdates)

        snapshot = json.load(open(snapshot_file))
        await shapper.on_snaphsot_async(snapshot)


        file_updates = tqdm(self.book_updates_and_trades(pair))
        for fupdate, ftrades in file_updates:
                file_updates.set_description(fupdate.replace(self.data_folder, ''))
                async with aiofiles.open(ftrades, 'r') as fp:
                    list_trades = json.loads(await fp.read())
                    await shapper.on_trades_bunch(list_trades)
                js = json.load(open(fupdate))
                allupdates = tqdm(js, leave=False)

                for book_upd in allupdates:
                    if book_upd['e'] != 'depthUpdate':
                        print("not update:", book_upd['e'])
                        continue
                    firstID = book_upd['U']
                    finalID = book_upd['u']
                    eventTime = book_upd['E']
                    ts = 1 + eventTime // 1000

                    if next_snap and finalID >= next_snap:
                        snapshot = json.load(open(snapshot_file))

                        await shapper.on_snaphsot_async(snapshot)
                        try:
                            next_snap,snapshot_file = next(snapshotupdates)
                        except StopIteration as e:
                            next_snap = None

                    px = await shapper.on_depth_msg_async(book_upd)

                    t_avail = shapper.secondAvail(book_upd)
                    oneSec = await shapper.make_frames_async(t_avail)
                    BBO = oneSec['bids'].index[0], oneSec['asks'].index[0]

                    allupdates.set_description(f"ts={datetime.datetime.utcfromtimestamp(ts)}, tr={len(oneSec['trades']):02}, BBO:{BBO}")#", px={px:16.12f}")
                    # assert not prev_ts or ts == 1 + prev_ts
                    # prev_ts = ts

                    yield oneSec

    @staticmethod
    async def multireplayL2_async(replayers):
        pairs = [await replayer.__anext__() for replayer in replayers]
        gens = {pairs[i]: replayers[i] for i in range(len(pairs))}
        nexs = {pair: await gens[pair].__anext__() for pair in pairs}
        def next_sec(pair): return nexs[pair]['time']
        tall = max([next_sec(p) for p in pairs])
        curs = {pair: nexs[pair] for pair in pairs}
        print('tall', tall)
        while True:
            for pair in pairs:
                while next_sec(pair) < tall:
                    curs[pair] = nexs[pair]
                    try:
                        nexs[pair] = await gens[pair].__anext__()
                    except StopAsyncIteration:
                        return # break
            yield curs
            next_overall_sec = min([next_sec(pair) for pair in pairs])
            jump = next_overall_sec - tall
            if jump > 60:
                print(f"\njumping {datetime.timedelta(seconds=jump)} seconds to have an update from one of the symbols")
                tall += jump
            else:
                tall += 1


async def main():
    markets = ['ETHBTC']
    file_replayer = Replayer('../data/crypto')
    s = file_replayer.zipped()
    print(s)

    return
    markets = ['ETHBTC']
    file_replayer = Replayer('../crypto-trading/data/L2')
    replay = file_replayer.replayL2('ETHBTC', await BookShapper.create())
    for d in replay:
        pass
        break
    
    replay = file_replayer.replayL2_async('ETHBTC', await BookShapper.create())
    async for d in replay:
        pass
        break

    replayers = [file_replayer.replayL2_async(pair, await BookShapper.create()) for pair in markets]
    multi_replay = file_replayer.multireplayL2_async(replayers)
    async for d in multi_replay:
        pass

if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



from deep_orderbook.recorder import Receiver, Writer
from deep_orderbook.shapper import BookShapper

import asyncio
import pandas as pd
import numpy as np
from pylab import rcParams
import aioitertools


MARKETS = ["ETHBTC", "BTCUSDT", "ETHUSDT", "BNBBTC", "BNBETH", "BNBUSDT"]


class IamgeStream:
    def __init__(self, src):
        self.frame = None
        self.markets = src
    
    async def setup(self):
        self.receiver = await Receiver.create(markets=self.markets, print_level=1)
        #self.receiver = await Writer.create(markets=MARKETS, data_folder='../crypto-trading/data')
        #wrting = asyncio.create_task(self.receiver.run_writer(save_period_minutes=10))


        shappers = {pair: await BookShapper.create() for pair in self.markets}
        multi_replay = self.receiver.multi_generator(shappers)


        _ = await multi_replay.__anext__()


        genarr = BookShapper.gen_array_async(market_replay=multi_replay, markets=self.markets)
        _ = await aioitertools.next(genarr)


        self.genacc = BookShapper.accumulate_array(genarr, markets=self.markets)
        _ = await aioitertools.next(self.genacc)

    async def run(self):
        async for toshow in BookShapper.images(accumulated_arrays=self.genacc, every=1, LENGTH=512):
            self.frame = toshow.copy()

    def start(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
#        loop.run_until_complete(do_stuff(i))
#        loop.close()
        asyncio.get_event_loop().run_until_complete(self.setup())
        asyncio.get_event_loop().run_until_complete(self.run())

    def stop(self):
        pass

    def read(self):
        return self.frame if self.frame is None else self.frame * 255


import json
import itertools
import collections
import time, datetime
import pandas as pd
import numpy as np
import asyncio
import aiofiles

from tqdm.auto import tqdm
from deep_orderbook.recorder import MessageDepthCacheManager
import aioitertools

pd.set_option('precision', 12)
import matplotlib.pyplot as plt


class BookShapper:
    PriceShape = [2,3]

    @classmethod
    async def create(cls):
        self = cls()
        self._depth_manager = await MessageDepthCacheManager.create(client=None, loop=None, symbol=None, refresh_interval=None)
        self.sec_trades = dict()
        self.bids = None
        self.asks = None
        self.tpr = None
        self.ts = None
        self.px = None
#        self.prev_px = None
        self.emaPrice = None
        self.emaNew = 1/32
        self.emptyframe = pd.DataFrame(columns=['p', 'q', 'delay', 'num', 'up']).set_index(['p'])
        return self

    async def on_snaphsot_async(self, snapshot):
        await self._depth_manager._init_cache(snapshot)

    async def on_depth_msg_async(self, msg):
        await self._depth_manager._depth_event(msg)
        bids, asks = self._depth_manager.get_depth_cache().get_bids_asks()
        ts = self.secondAvail(msg)
        await self.update_ema(bids, asks, ts)

    async def update_ema(self, bids, asks, ts):
        self.ts = ts
        bbp, bbs = bids[0]
        bap, bas = asks[0]
        price = (bbp * bas + bap * bbs) / (bbs + bas)
        self.px = round(price, 8)
        self.emaPrice = self.px * self.emaNew + (self.emaPrice if self.emaPrice is not None else self.px) * (1-self.emaNew)
        self.px += 1e-12
        return self.px

    @staticmethod
    def secondAvail(tr_dict):
        return 1 + tr_dict['E'] // 1000

    async def on_trades_bunch(self, list_trades, force_t_avail=None):
        list_trades = [{k:float(v) for k,v in trs.items() if k not in 'Msea'} for trs in list_trades]
        if not list_trades:
            return
        if force_t_avail:
            self.sec_trades[force_t_avail] = self.trades2frame(list_trades).drop(['tavail'], axis=1)
            return
        alltrades = self.trades2frame(list_trades)
        for i,l in alltrades.groupby('tavail'):
            # print(f'{i}\n{l}')
            self.sec_trades[i] = l.drop(['tavail'], axis=1)
        return

    @staticmethod
    def trades2frame(list_trades):
        ts = pd.DataFrame(list_trades).set_index('p')
        ts['tavail'] = ts['E'] // 1000 + 1
        ts['delay'] = ts['E'] - ts['T']
        ts['num'] = ts['l'] - ts['f'] + 1
        ts['up'] = 1 - 2*ts['m']
        ts = ts.drop(['E', 'T', 'f', 'l', 'm'], axis=1)

        #ts.loc[self.px] = 0
        ts.sort_index(inplace=True)
#        self.prev_px = self.px
        return ts

    async def make_frames_async(self, t_avail, bids=None, asks=None):
        if bids is None or asks is None:
            bids, asks = self._depth_manager.get_depth_cache().get_bids_asks()

        oneSec = {'time': self.ts,
                  'price': self.px,
                  'bids': pd.DataFrame(bids, columns=['price', 'size']).set_index('price'),
                  'asks': pd.DataFrame(asks, columns=['price', 'size']).set_index('price'),
                  'trades': self.sec_trades.pop(t_avail, self.emptyframe),
                 'emaPrice': self.emaPrice,
                 }
        return oneSec





    FRAC_LEVELS = 0.01
    NUM_LEVEL_BINS = 128
    SPACING = np.cumsum(0+np.linspace(0, NUM_LEVEL_BINS, NUM_LEVEL_BINS, endpoint=False))# * 4
    SPACING = SPACING / SPACING[-1]
    @staticmethod
    def bin_books(dfb, dfa, tr, ref_price, zoom_frac=FRAC_LEVELS, spacing=SPACING):
        b_idx = np.round(pd.Index(ref_price * (1-spacing*zoom_frac)), 7)
        a_idx = np.round(pd.Index(ref_price * (1+spacing*zoom_frac)), 7)

        t_idx = b_idx[::-1].append(a_idx)
        t_idx_inv = t_idx[::-1]
        
        reind_b = dfb.cumsum().reindex(t_idx_inv, method='ffill', fill_value=0).diff().fillna(0)[::-1]
        reind_a = dfa.cumsum().reindex(t_idx, method='ffill', fill_value=0).diff().fillna(0)
        treind_b = tr[tr['up']<=0].groupby(level=0).sum()[::-1].cumsum().reindex(t_idx_inv, method='ffill', fill_value=0).diff().fillna(0)[::-1]
        treind_a = tr[tr['up']>=0].groupby(level=0).sum().cumsum().reindex(t_idx, method='ffill', fill_value=0).diff().fillna(0)
        
        treind_b = np.arcsinh(treind_b).astype(np.float32)
        treind_a = np.arcsinh(treind_a).astype(np.float32)
        reind_b = np.arcsinh(reind_b).astype(np.float32)
        reind_a = np.arcsinh(reind_a).astype(np.float32)
        return reind_b, reind_a, treind_b, treind_a



    def sampleArrays(self, replayer, numpoints=None, apply_fnct=None):
        arrs = []
        trrs = []
        pric = []
        prev_price = None
        i = 0
        spacing = np.arange(64)
        #spacing = np.square(spacing) + spacing
        spacing = spacing / spacing[-1]
        for bi,ai,tpi,tri in replayer:
            prev_price = prev_price or tpi['price']
            bib,aib,trb,tra = self.bin_books(bi,ai,tri, ref_price=prev_price, zoom_frac=1/256, spacing=spacing)
            prev_price = tpi['emaPrice']#0.5*(tpi['bid'] + tpi['ask'])
            arrs.append(np.concatenate([bib.values - aib.values]))
            trrs.append(np.concatenate([trb.values - tra.values]))
            pric.append(np.array([tpi['price'], tpi['emaPrice'], tpi['bid'], tpi['ask'], tpi['time']]))
            i += 1
            if apply_fnct and i%10 == 0:
                apply_fnct(**{'books': np.stack(arrs), 'prices': np.stack(pric), 'trades':np.stack(trrs)})
            if numpoints and i >= numpoints:
                break
        books = np.stack(arrs)
        prices = np.stack(pric)
        trades = np.stack(trrs)
        ret = {'books': books, 'prices': prices, 'trades':trades}
        return ret

    def sampleImages(self, books, prices, trades):
        #print(books.shape, prices.shape, trades.shape)
        plt.margins(0.0)
        plt.plot(prices[:, 0])
        plt.plot(prices[:, 1])
        plt.plot(prices[:, 2])
        plt.plot(prices[:, 3])
        plt.show()
        im = np.abs(books[:, :, 0]).copy()
        im[im == 0] = -1
        plt.imshow(im.T, cmap='nipy_spectral', origin="lower")
        plt.show()
        im = np.abs(trades[:, :, 2])
        #im[im == 0] = 3
        plt.imshow(im.T, cmap='nipy_spectral', origin="lower")
        plt.show()
        im0 = books[:, :, 0].T/10
        im1 = trades[:, :, 0].T/1
        im2 = trades[:, :, 2].T/1
        im3 = np.stack([im0, im1, im2], -1)+0.5
        print(im3.shape)
        plt.imshow(im3[:,:,:], origin="lower")
        plt.show()

    @staticmethod
    async def gen_array_async(market_replay, markets, width_per_side=64, zoom_frac=1/256):
        #market_replay = self.multireplayL2(markets)
        prev_price = {p: None for p in markets}
        spacing = np.arange(width_per_side)
        #spacing = np.square(spacing) + spacing
        spacing = spacing / spacing[-1]
        spacing = np.arcsin(spacing)*3 - spacing*2
        async for second in market_replay:
            market_second = {}#collections.defaultdict(list)
            for pair in markets:
                sec = second[pair]
                prev_price[pair] = prev_price[pair] or sec['price']
                bib,aib,trb,tra  = BookShapper.bin_books(sec['bids'],sec['asks'],sec['trades'], ref_price=prev_price[pair], zoom_frac=zoom_frac, spacing=spacing)
                prev_price[pair] = sec['emaPrice']
                arr0 = bib.values - aib.values
                arr1 = tra.values - trb.values
                utc = datetime.datetime.utcfromtimestamp(sec['time'])
                d = sec['time'] // (3600 * 24) # int(utc.strftime('%y%m%d'))
                t = sec['time'] % (3600 * 24)  #float(utc.strftime('%H%M%S.%f'))
                lowtrade = sec['trades'].index.min()
                hightrade = sec['trades'].index.max()
                #print(lowtrade, hightrade)
                tp = np.array([[lowtrade, sec['bids'].index[0], sec['asks'].index[0]], [d, t, hightrade]], dtype=np.float32)
                # print('tp', tp)
                arr3d = np.concatenate([arr0, arr1[:,::2]], axis=-1)
                market_second[pair] = {'ps': [tp], 'bs': [arr3d]}
            yield market_second


    @staticmethod
    def build_time_level_trade(books, prices, sidebips=32, sidesteps=64):
        pricestep = prices[0, 0, 0] * 0.0001 * sidebips / sidesteps
        FUTURE = 120#0*10
        time2levels = np.zeros_like(books[:, :2*sidesteps, :1]) + FUTURE
        #########################################################
                                    #######
        for i in tqdm(range(prices.shape[0])):
                                    #######
        #########################################################
            timeupdn = []
            for j in range(sidesteps):
                thresh = j * pricestep
                [_, b, a], [d, t, _] = prices[i]
                bids = prices[i:i+FUTURE, 0, 1]
                asks = prices[i:i+FUTURE, 0, 2]
                waitUp = bids < a + thresh
                waitDn = asks > b - thresh
                #### trades also define potential price hit
                lowtrade = prices[i:i+FUTURE, 0, 0]
                hightrade = prices[i:i+FUTURE, 1, 2] 
                tradeUp = hightrade >= a + thresh
                tradeDn = lowtrade <= b - thresh
                # cannot trade with bid/ask of elapsed second
                tradeUp[0] = False
                tradeDn[0] = False
                # first one to NOT wait defines the price hit
                waitUp &= ~tradeUp
                waitDn &= ~tradeDn
                ###########################################
                timeUp = np.argmin(waitUp) or FUTURE*10
                timeDn = np.argmin(waitDn) or FUTURE*10
                timeupdn.insert(0, [timeDn])
                timeupdn.append([timeUp])
            time2levels[i] = timeupdn
        return time2levels.astype(np.float32)


    @staticmethod
    def build(total, element, reduce_func=None):
        force_save = element is None
        element = element or total
        for market,second in element.items():
            dt = total[market]['ps'][-1][1] # datetime.datetime.strptime(str(int(total[market]['ps'][-1][1,0])), '%y%m%d').date()
            datetotal = datetime.datetime.utcfromtimestamp(int(dt[0]) * 3600 * 24 + int(dt[1])).date()
            de = element[market]['ps'][-1][1] # dateeleme = datetime.datetime.strptime(str(int(element[market]['ps'][-1][1,0])), '%y%m%d').date()
            # print(dt, dt[0] * 3600 * 24 + dt[1])
            dateeleme = datetime.datetime.utcfromtimestamp(int(de[0]) * 3600 * 24 + int(de[1])).date()
            newDay = datetotal < dateeleme
            if not (newDay or force_save):
                for name,arrs in second.items():
                    total[market][name] += arrs
            else:
                arrday_bs = np.stack(total[market]['bs']).astype(np.float32)
                np.save(f'data/{datetotal}-{market}-bs.npy', arrday_bs)
                arrday_ps = np.stack(total[market]['ps']).astype(np.float32)
                np.save(f'data/{datetotal}-{market}-ps.npy', arrday_ps)
                # np.save(f'data/{datetotal}-{market}-ptbs.npy', np.concatenate([arrday_ps, arrday_bs], axis=1))
                if reduce_func is not None:
                    t2l = reduce_func(
                        np.stack(total[market]['bs']).astype(np.float32),
                        np.stack(total[market]['ps']).astype(np.float32),
                    )
                    np.save(f'data/{datetotal}-{market}-time2level.npy', t2l)
            if newDay:
                for name,arrs in second.items():
                    total[market][name] = arrs
        return total

    @staticmethod
    def accumulate_array(genarr, markets):
        genacc = aioitertools.accumulate(genarr, BookShapper.build)
        return genacc

    @staticmethod
    async def images(accumulated_arrays, every=10, LENGTH=128):
        async for n,sec in aioitertools.enumerate(accumulated_arrays):
            if n % every:
                continue
            allim = []
            for symb, data in sec.items():
                arr = np.stack(data['bs'][-LENGTH:])
                im = arr
                im[:,:,0] /= 10
                im += 0.5
                im = im.transpose(1,0,2)
                im = np.clip(im, 0, 1)
                allim.append(im[::-1])
            toshow = np.concatenate(allim, axis=0)
            yield toshow


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    shapper = loop.run_until_complete(BookShapper.create())



import time
import datetime
import json
import os, sys
import collections
import copy
import threading
import asyncio
import aiofiles
from binance import AsyncClient, DepthCacheManager # Import the Binance Client

# Import the Binance Socket Manager
from binance.websockets import BinanceSocketManager
from binance.exceptions import BinanceAPIException
from binance.depthcache import DepthCache

# https://github.com/binance-exchange/binance-official-api-docs/blob/master/web-socket-streams.md#how-to-manage-a-local-order-book-correctly
DEBUG = False

class DepthCachePlus(DepthCache):
    def add_bid(self, bid):
        pr = float(bid[0])
        sz = float(bid[1])
        self._bids[pr] = sz
        if sz == 0.0:
            del self._bids[pr]

    def add_ask(self, ask):
        pr = float(ask[0])
        sz = float(ask[1])
        self._asks[pr] = sz
        if sz == 0.0:
            del self._asks[pr]

    def get_bids_asks(self):
        bids = self.get_bids()
        asks = self.get_asks()
        if bids[0][0] >= asks[0][0]:
            print(f"\ncleaning the crossed BBO \nBIDS: {bids[:5]}\nASKS: {asks[:5]}")
            for p in list(self._bids.keys()):
                if p >= asks[0][0]:
                    if DEBUG:
                        print(f"del bids[{p}]")
                    del self._bids[p]
            for p in list(self._asks.keys()):
                if p <= bids[0][0]:
                    if DEBUG:
                        print(f"del asks[{p}]")
                    del self._asks[p]
            bids = self.get_bids()
            asks = self.get_asks()
            print(f"result: \nBIDS: {bids[:5]}\nASKS: {asks[:5]}")

        assert bids[0][0] < asks[0][0]
        return bids, asks


class MessageDepthCacheManager(DepthCacheManager):
    _default_refresh = 60 * 30  # 30 minutes
    @classmethod
    async def create(cls, client, loop, symbol, coro=None, refresh_interval=_default_refresh, bm=None, limit=500, msg_coro=None):
        self = MessageDepthCacheManager()
        self._client = client
        self._loop = loop
        self._symbol = symbol
        self._limit = limit
        self._coro = coro
        self._msg_coro = msg_coro
        self._last_update_id = None
        self._depth_message_buffer = []
        self._bm = bm
        self._depth_cache = DepthCachePlus(self._symbol)
        self._refresh_interval = refresh_interval
        self.trades = list()

        if self._client:
            await self._start_socket()
            await self._init_cache()

        return self
        
    async def _init_cache(self, snapshot=None):
        if self._client:
            await super()._init_cache()
            return

        if not snapshot:
            return

        self._last_update_id = None
        self._depth_message_buffer = []
        res = snapshot
        self._depth_cache._bids = {}
        self._depth_cache._asks = {}

        # process bid and asks from the order book
        for bid in res['bids']:
            self._depth_cache.add_bid(bid)
        for ask in res['asks']:
            self._depth_cache.add_ask(ask)

        # set first update id
        self._last_update_id = res['lastUpdateId']
        # print("\nfirst update", self._last_update_id)

#        # set a time to refresh the depth cache
#        if self._refresh_interval:
#            self._refresh_time = int(time.time()) + self._refresh_interval

        # Apply any updates from the websocket
        for msg in self._depth_message_buffer:
            await self._process_depth_message(msg, buffer=True)

        # clear the depth buffer
        del self._depth_message_buffer

    async def _depth_event(self, msg):
        await super()._depth_event(msg)
        if self._msg_coro:
            await self._msg_coro(msg)

class Receiver:

    @classmethod
    async def create(cls, **kwargs):
        # print(cls)
        self = cls()
        await self.setup(**kwargs)
        return self

    async def setup(self, markets, print_level=2):
        self.last_update_time = time.time()
        self.print_level = print_level
        self.markets = markets
        # Instantiate a Client
        self.client = await AsyncClient.create()
        #print(json.dumps(await self.client.get_exchange_info(), indent=2))
        for m in self.markets:
            print(json.dumps(await self.client.get_symbol_ticker(symbol=m), indent=2))

        # Instantiate a BinanceSocketManager, passing in the client that you instantiated
        self.bm = BinanceSocketManager(self.client, loop=asyncio.get_event_loop())
        self.nummsg = collections.defaultdict(int)
        self.conn_keys = []
        self.depth_managers = {}
        self.trade_managers = collections.defaultdict(list)

        await self.stoprestart()

    async def on_depth_msg(self, msg):
        """
        {
            "e": "depthUpdate", # Event type
            "E": 123456789,     # Event time
            "s": "BNBBTC",      # Symbol
            "U": 157,           # First update ID in event
            "u": 160,           # Final update ID in event
            "b": [              # Bids to be updated
                [
                    "0.0024",   # price level to be updated
                    "10",       # quantity
                    []          # ignore
                ]
            ],
            "a": [              # Asks to be updated
                [
                    "0.0026",   # price level to be updated
                    "100",      # quantity
                    []          # ignore
                ]
            ]
        }
        """
        symbol = msg['s']

    async def on_depth(self, depth_cache):
        symbol = depth_cache.symbol
        self.depth_managers[symbol].trades = copy.deepcopy(self.trade_managers[symbol])
        self.trade_managers[symbol] = list()

    async def on_aggtrades(self, msg):
        """
        {
            "e": "aggTrade",                # event type
            "E": 1499405254326,             # event time
            "s": "ETHBTC",                  # symbol
            "a": 70232,                             # aggregated tradeid
            "p": "0.10281118",              # price
            "q": "8.15632997",              # quantity
            "f": 77489,                             # first breakdown trade id
            "l": 77489,                             # last breakdown trade id
            "T": 1499405254324,             # trade time
            "m": false,                             # whether buyer is a maker
            "M": true                               # can be ignored
        }
        """
        self.last_update_time = time.time()
        symbol = msg["s"]
        self.trade_managers[symbol].append(copy.deepcopy(msg))
        self.nummsg[symbol] += 1
        if self.print_level >= 2:
            print(', '.join([f'{s}: {self.nummsg[s]:06}' for s in self.markets]), end='\r')

    async def stoprestart(self, dorestart=True):
        # stop the socket manager
        for conn_key in self.conn_keys:
            print(f"stopping socket {conn_key}\n")
            #await self.bm.stop_socket(conn_key)
        conn_keys = []
        self.depth_managers = {}
#        await self.bm.close()

        if dorestart:
            for symbol in self.markets:
                key = await self.bm.start_aggtrade_socket(symbol, self.on_aggtrades)
                print("start", key)
                conn_keys.append(key)
            # create the Depth Cache
            for symbol in self.markets:
                depthmanager = await MessageDepthCacheManager.create(self.client,
                                                                     asyncio.get_event_loop(), 
                                                                     symbol, 
                                                                     self.on_depth,
                                                                     bm=self.bm,
                                                                     limit=1000,
                                                                     msg_coro=self.on_depth_msg
                                                                     )
                self.depth_managers[symbol] = depthmanager

    async def multi_generator(self, symbol_shappers):
        tall = time.time()
        tall = tall // 1
        while True:
            twake = tall
            timesleep = twake - time.time()
            if timesleep > 0:
                await asyncio.sleep(timesleep)
            else:
                print(f"time sleep is negative: {timesleep}")

            oneSec = {}
            for symbol,shapper in symbol_shappers.items():
                bids, asks = self.depth_managers[symbol].get_depth_cache().get_bids_asks()
                await shapper.update_ema(bids, asks, twake)
                list_trades = self.depth_managers[symbol].trades
                await shapper.on_trades_bunch(list_trades, force_t_avail=twake)
                oneSec[symbol] = await shapper.make_frames_async(t_avail=twake, bids=bids, asks=asks)
            yield oneSec
            tall += 1


class Writer(Receiver):
    async def setup(self, markets, data_folder):
        self.store = collections.defaultdict(list)
        self.tradestore = collections.defaultdict(list)
        self.L2folder = f"{data_folder}/L2"
        self.lock = threading.Lock()
        self.tradelock = threading.Lock()

        for symbol in markets:
            os.makedirs(f"{self.L2folder}/{symbol}", exist_ok=True)

        await super().setup(markets)

    async def on_depth_msg(self, msg):
        await super().on_depth_msg(msg)
        symbol = msg['s']
        with self.lock:
            self.store[symbol].append(msg)

    async def on_depth(self, depth_cache):
        await super().on_depth(depth_cache)
        symbol = depth_cache.symbol
        with self.tradelock:
            self.tradestore[symbol] += self.depth_managers[symbol].trades

    async def save_updates_since(self, prev_ts=None):
        prev_ts = prev_ts or self.prev_th
        progressbar = self.markets
        upds = datetime.datetime.utcfromtimestamp(prev_ts).isoformat().replace(":", "-")  # .replace('-',"_")
        for symbol in progressbar:
            with self.lock:
                tosave = copy.deepcopy(self.store[symbol])
                self.store[symbol] = list()
                self.nummsg[symbol] = 0
            with self.tradelock:
                tradetosave = copy.deepcopy(self.tradestore[symbol])
                self.tradestore[symbol] = list()

            async with aiofiles.open(f"{self.L2folder}/{symbol}/{upds}_update.json", "w") as fp:
                await fp.write(json.dumps(tosave))
            async with aiofiles.open(f"{self.L2folder}/{symbol}/{upds}_trades.json", "w") as fp:
                await fp.write(json.dumps(tradetosave))
        print(f"\nsaved_updates since {upds}")

    async def save_snapshot(self, cur_ts, max_levels=1000):
        snap = datetime.datetime.utcfromtimestamp(cur_ts).isoformat().replace(":", "-")  # .replace('-',"_")
        if cur_ts:
            L2s_coro = [self.client.get_order_book(symbol=pair, limit=max_levels) for pair in self.markets]
            L2s = await asyncio.gather(*L2s_coro)
            for symbol, L2 in zip(self.markets, L2s):
                async with aiofiles.open(f"{self.L2folder}/{symbol}/{snap}_snapshot.json", "w") as fp:
                    await fp.write(json.dumps(L2))
        print("\nsaved_snapshot")

    async def run_writer(self, save_period_minutes=60):
        save_period_seconds = save_period_minutes * 60

        PERIOD_L2 = 10
        self.th = 0
        self.prev_th = 0

        try:
            while True:
                t = time.time()
                await asyncio.sleep(PERIOD_L2 - t % PERIOD_L2)
                t_ini = int(time.time())
                self.new_th = int(t_ini // save_period_seconds) * save_period_seconds

                if self.new_th > self.th:
                    if self.th == 0:
                        await self.save_updates_since(t_ini)
                        await self.save_snapshot(t_ini)
                        self.prev_th = t_ini
                    else:
                        await self.save_updates_since()
                        await self.save_snapshot(self.new_th)
                        self.prev_th = self.new_th
                    self.th = self.new_th
        except asyncio.CancelledError as e:
            await self.save_updates_since()
        except Exception as e:
            print(e.__class__, e)
            raise e


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    MARKETS = ["BNBUSDT", "BTCUSDT", "ETHUSDT", "BNBBTC", "ETHBTC", "BNBETH"]
    rec = loop.run_until_complete(Receiver.create(markets=MARKETS))#['ETHBTC', 'BTCUSDT']))
    while True:
        loop.run_until_complete(asyncio.sleep(10))



import asyncio
import json

from binance import AsyncClient, DepthCacheManager, BinanceSocketManager


async def main():
    # initialise the client
    client = await AsyncClient.create()

    # run some simple requests
    print(json.dumps(await client.get_exchange_info(), indent=2))

    print(json.dumps(await client.get_symbol_ticker(symbol="BTCUSDT"), indent=2))


    # initialise socket manager
    bsm = BinanceSocketManager(client, asyncio.get_event_loop())

    # setup async callback handler for socket messages
    async def handle_evt(msg):
        pair = msg['s']
        print(f'{pair} : {msg}')

    # create listener, can use the `ethkey` value to close the socket later
    trxkey = await bsm.start_trade_socket('TRXBTC', handle_evt)


    # setup an async callback for the Depth Cache
    async def process_depth(depth_cache):
        print(f"symbol {depth_cache.symbol} updated:{depth_cache.update_time}")
        print("Top 5 asks:")
        print(depth_cache.get_asks()[:5])
        print("Top 5 bids:")
        print(depth_cache.get_bids()[:5])

    # create the Depth Cache
    dcm1 = await DepthCacheManager.create(client, asyncio.get_event_loop(), 'TRXBTC', process_depth)

    while True:
        print("doing a sleep")
        await asyncio.sleep(20)


if __name__ == "__main__":

    asyncio.get_event_loop().run_until_complete(main())
    


import runpy
from setuptools import setup, find_packages

PACKAGE_NAME = "deep_orderbook"
version_meta = runpy.run_path("./version.py")
VERSION = version_meta["__version__"]


with open("README.md", "r") as fh:
    long_description = fh.read()


def parse_requirements(requirements_file='requirements.txt'):
    """Get the contents of a file listing the requirements"""
    lines = open(requirements_file).readlines()
    dependencies = []
    for line in lines:
        maybe_dep = line.strip()
        if maybe_dep.startswith('#'):
            # Skip pure comment lines
            continue
        if maybe_dep.startswith('git+'):
            # VCS reference for dev purposes, expect a trailing comment
            # with the normal requirement
            __, __, maybe_dep = maybe_dep.rpartition('#')
        else:
            # Ignore any trailing comment
            maybe_dep, __, __ = maybe_dep.partition('#')
        # Remove any whitespace and assume non-empty results are dependencies
        maybe_dep = maybe_dep.strip()
        if maybe_dep:
            dependencies.append(maybe_dep)
    return dependencies

if __name__ == "__main__":
    setup(
        name=PACKAGE_NAME,
        version=VERSION,
        description="Transforms orders books in temporally and spatially local-correlated images.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/gQuantCoder/deep_orderbook",
        author="gQuantCoder",
        author_email="gquantcoder@gmail.com",
        packages=find_packages(),
        install_requires=parse_requirements("requirements.txt"),
        python_requires=">=3.7.4",
        entry_points={"console_scripts": ["deepbook=deep_orderbook.__main__:main"]},
    )